@@ -1,11 +1,10 @@
 from __future__ import absolute_import, division, print_function
 
-from .core import Evaluation
+from .core import Evaluation, setup
 from . import core
 from . import scheme
 from . import lut
 from . import models
-from . import progressbar
 from . import util
 
 from pkg_resources import get_distribution
